(prompt "
  xreplace.lsp - switches block insertions to another definition
  10/2003 knut hohenberg, no rights reserved."
)
;4/2017: added automatic unit scaling
;3/2019: added global replacement option


(defun set_block (val / )
 ;(setq srcnme (cons 2 (nth (atoi val) blklst)))
 (setq srcnme (nth (atoi val) blklst))
)

(defun get_block_units (blknme / bit70)
 (vlax-for x
  (vla-get-Blocks (vla-get-activedocument (vlax-get-acad-object)))
  (if (= (vla-get-name x) blknme)
   ;(vlax-get-property x "Units") ; does not work for xrefs
   (setq bit70 (cdr (assoc 70 (entget (vlax-vla-object->ename x)))))
  )
 )
 (eval bit70)
)

(defun unit->mm (unit / )
 (cond
  ((= 1 unit) 25.4)
  ((= 2 unit) 304.8)
  ((= 3 unit) 1609344.0)
  ((= 4 unit) 1.0)
  ((= 5 unit) 10.0)
  ((= 6 unit) 1000.0)
  ((= 7 unit) 1000000.0)
  (t 0)
 )
)

(defun C:xreplace ( / entnme source srcnme srcpth srcmm pstrng istrng global insmm insscl dcl_id n blknme blklst target insdat selset)
 (setq *error* bs_err)
 (setq entnme nil)
 (initget "File Block")
 (setq entnme (entsel "\nSelect an insert of the block definition to switch to or choose [File/Block]<Block>: "))
 (cond
  ((or (not entnme) (= "Block" entnme))
   (setq dcl_id (load_dialog "xreplace.dcl"))
   (if (not (new_dialog "xreplace" dcl_id )) (exit))
   (setq blklst '())
   (setq n nil)
   (while (setq n (tblnext "BLOCK" (not n)))
    (if (/= (logand 1 (cdr (assoc 70 n))) 1)
     (progn
      (setq blknme (cdr (assoc 2 n)))
      (if (and (/= "*" (substr blknme 1 1))(not(wcmatch blknme "*|*")))
       (setq blklst (cons blknme blklst))
      )
     )
    )
   )
   (setq blklst (acad_strlsort blklst))
   ;(setq srcnme (cons 2 (car blklst)))
   (setq srcnme (car blklst))
   (start_list "blocks")
   (mapcar 'add_list blklst)
   (end_list)
   (if (/= 1 (start_dialog))
    (exit)
   )
   (unload_dialog dcl_id)
   (setq entnme nil)
  )
  ((= "File" entnme)
   (setq srcpth (getfiled "Select Drawing" (getvar "dwgprefix") "dwg" 0))
   (if (not srcpth) (exit))
   (setq srcnme (vl-string-subst "" ".dwg" (acet-filename-path-remove srcpth)))
   (print srcnme)
   (if (tblsearch "BLOCK" srcnme)
    (progn
     (alert (strcat "\nA block named " srcnme " is already defined in the drawing.\nYou have to rename the block prior to replacement."))
     (exit)
    )
   )
   (initget "Block Xref")
   (if (= "Xref" (getkword "\nInsertion mode [as Block/as Xref] <as Block>: "))
    (progn
     (entmake (list '(0 . "BLOCK") (cons 2  srcnme) '(70 . 4) '(10 0.0 0.0 0.0) (cons 1  srcpth)))
     (entmake '((0 . "ENDBLK")))
    )
    (progn
     (command "_.-Insert" srcpth)
     (command)
    )
   )
   (setq entnme nil)
  ) 
  (t
   (setq entnme (car entnme))
   (redraw entnme 3)
   (setq source (entget entnme))
   (if (= "INSERT" (cdr (assoc 0 source)))
    (setq srcnme (cdr (assoc 2 source)))
    (progn
     (redraw entnme 4)
     (prompt "No Insert found.")
     (exit)
    )
   )
  )
 )
 
 (setq srcmm (unit->mm (get_block_units srcnme)))
 (setq global nil)
 (setq pstrng "\nSelect insertion to replace or [switch to Global mode]: ")
 (setq istrng "Global")
 (initget istrng)
 (while (setq target (entsel pstrng))
  (cond
   ((= target "Global")
    (prompt "\nSwitched to global mode")
    (setq pstrng "\nSelect insertion for global replacement or [switch to Individual mode]: ")
    (setq global t)
    (setq istrng "Individual")
    (initget istrng)
   )
   ((= target "Individual")
    (prompt "\nSwitched to individual mode")
    (setq pstrng "\nSelect insertion to replace or [switch to Global mode]: ")
    (setq global nil)
    (setq istrng "Global")
    (initget istrng)
   )
   (t
    (setq target (car target))
    (setq insdat (entget target))
    (setq insmm (unit->mm (get_block_units (cdr (assoc 2 insdat)))))
    (if (and (/= 0 srcmm)(/= 0 insmm) (/= insmm srcmm))
     (setq insscl (/ srcmm insmm))
     (setq insscl 1.0)
    )
    (cond
     (global
      (setq selset (ssget "X" (list '(0 . "INSERT") (cons 2 (cdr (assoc 2 insdat))))))
      (repeat (setq n (sslength selset))
       (setq insdat (entget (ssname selset (setq n (1- n))))
             insdat (subst (cons 41 (* insscl (cdr (assoc 41 insdat)))) (assoc 41 insdat) insdat)
             insdat (subst (cons 42 (* insscl (cdr (assoc 42 insdat)))) (assoc 42 insdat) insdat)
             insdat (subst (cons 43 (* insscl (cdr (assoc 43 insdat)))) (assoc 43 insdat) insdat)
       )
       (setq insdat (subst (cons 2 srcnme) (assoc 2 insdat) insdat))
       (entmod insdat)
      )
     )
     (t
      (setq 
            insdat (subst (cons 41 (* insscl (cdr (assoc 41 insdat)))) (assoc 41 insdat) insdat)
            insdat (subst (cons 42 (* insscl (cdr (assoc 42 insdat)))) (assoc 42 insdat) insdat)
            insdat (subst (cons 43 (* insscl (cdr (assoc 43 insdat)))) (assoc 43 insdat) insdat)
            insdat (subst (cons 2 srcnme) (assoc 2 insdat) insdat)
      )
      (entmod insdat)
     )
    )
    (initget istrng)
   )
  )
 )
 (setq *error* nil)
 (if entnme (redraw entnme 4))
 (princ)
)

(defun bs_err (msg)
  (if (/= msg "quit / exit abort")
   (princ (strcat "\nError: " msg))
  )
 (if entnme (redraw entnme 4))
 (setq *error* nil)
 (princ)
)


(prompt "
  ...loaded. Type XREPLACE to invoke.
"
)
(princ)
